# Copyright 2015 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import math
import operator
import string

from operator import methodcaller
from datetime import date, datetime

import pytest

import numpy as np

import ibis
import ibis.expr.types as ir
import ibis.expr.datatypes as dt
import ibis.config as config

from ibis import literal as L

import pandas as pd
import pandas.util.testing as tm

sa = pytest.importorskip('sqlalchemy')
pytest.importorskip('psycopg2')

pytestmark = pytest.mark.postgresql


@pytest.yield_fixture
def guid(con):
    name = ibis.util.guid()
    try:
        yield name
    finally:
        con.drop_table(name, force=True)


@pytest.yield_fixture
def guid2(con):
    name = ibis.util.guid()
    try:
        yield name
    finally:
        con.drop_table(name, force=True)


@pytest.mark.parametrize(
    ('left_func', 'right_func'),
    [
        (
            lambda t: t.double_col.cast('int8'),
            lambda at: sa.cast(at.c.double_col, sa.SMALLINT),
        ),
        (
            lambda t: t.double_col.cast('int16'),
            lambda at: sa.cast(at.c.double_col, sa.SMALLINT),
        ),
        (
            lambda t: t.string_col.cast('double'),
            lambda at: sa.cast(at.c.string_col, sa.FLOAT),
        ),
        (
            lambda t: t.string_col.cast('float'),
            lambda at: sa.cast(at.c.string_col, sa.REAL),
        ),
    ]
)
def test_cast(alltypes, at, translate, left_func, right_func):
    left = left_func(alltypes)
    right = right_func(at)
    assert str(translate(left).compile()) == str(right.compile())


@pytest.mark.xfail(raises=AssertionError, reason='NYI')
def test_decimal_cast():
    assert False


def test_date_cast(alltypes, at, translate):
    result = alltypes.date_string_col.cast('date')
    expected = sa.cast(at.c.date_string_col, sa.DATE)
    assert str(translate(result)) == str(expected)


@pytest.mark.parametrize(
    'column',
    [
        'index',
        'Unnamed: 0',
        'id',
        'bool_col',
        'tinyint_col',
        'smallint_col',
        'int_col',
        'bigint_col',
        'float_col',
        'double_col',
        'date_string_col',
        'string_col',
        'timestamp_col',
        'year',
        'month',
    ]
)
def test_noop_cast(alltypes, at, translate, column):
    col = alltypes[column]
    result = col.cast(col.type())
    expected = at.c[column]
    assert result.equals(col)
    assert str(translate(result)) == str(expected)


def test_timestamp_cast_noop(alltypes, at, translate):
    # See GH #592
    result1 = alltypes.timestamp_col.cast('timestamp')
    result2 = alltypes.int_col.cast('timestamp')

    assert isinstance(result1, ir.TimestampColumn)
    assert isinstance(result2, ir.TimestampColumn)

    expected1 = at.c.timestamp_col
    expected2 = sa.func.timezone('UTC', sa.func.to_timestamp(at.c.int_col))

    assert str(translate(result1)) == str(expected1)
    assert str(translate(result2)) == str(expected2)


@pytest.mark.parametrize(
    ('func', 'expected'),
    [
        (methodcaller('strftime', '%Y%m%d'), '20150901'),
        (methodcaller('year'), 2015),
        (methodcaller('month'), 9),
        (methodcaller('day'), 1),
        (methodcaller('hour'), 14),
        (methodcaller('minute'), 48),
        (methodcaller('second'), 5),
        (methodcaller('millisecond'), 359),
        (lambda x: x.day_of_week.index(), 3),
        (lambda x: x.day_of_week.full_name(), 'Tuesday'),
    ]
)
def test_simple_datetime_operations(con, func, expected, translate):
    value = ibis.timestamp('2015-09-01 14:48:05.359')
    assert con.execute(func(value)) == expected


@pytest.mark.parametrize(
    'func',
    [
        # there could be pathological failure at midnight somewhere, but
        # that's okay
        methodcaller('strftime', '%Y%m%d %H'),

        # test quoting behavior
        methodcaller('strftime', 'DD BAR %w FOO "DD"'),
        methodcaller('strftime', 'DD BAR %w FOO "D'),
        methodcaller('strftime', 'DD BAR "%w" FOO "D'),
        methodcaller('strftime', 'DD BAR "%d" FOO "D'),
        pytest.mark.xfail(
            methodcaller('strftime', 'DD BAR "%c" FOO "D'),
            condition=os.name == 'nt',
            reason='Locale-specific format specs not available on Windows',
        ),
        pytest.mark.xfail(
            methodcaller('strftime', 'DD BAR "%x" FOO "D'),
            condition=os.name == 'nt',
            reason='Locale-specific format specs not available on Windows',
        ),
        pytest.mark.xfail(
            methodcaller('strftime', 'DD BAR "%X" FOO "D'),
            condition=os.name == 'nt',
            reason='Locale-specific format specs not available on Windows',
        ),
    ]
)
def test_strftime(con, func):
    value = ibis.timestamp('2015-09-01 14:48:05.359')
    raw_value = datetime(
        year=2015, month=9, day=1,
        hour=14, minute=48, second=5, microsecond=359000
    )

    assert con.execute(func(value)) == func(raw_value)


@pytest.mark.parametrize(
    ('func', 'left', 'right', 'expected'),
    [
        (operator.add, L(3), L(4), 7),
        (operator.sub, L(3), L(4), -1),
        (operator.mul, L(3), L(4), 12),
        (operator.truediv, L(12), L(4), 3),
        (operator.pow, L(12), L(2), 144),
        (operator.mod, L(12), L(5), 2),
        (operator.truediv, L(7), L(2), 3.5),
        (operator.floordiv, L(7), L(2), 3),
        (lambda x, y: x.floordiv(y), L(7), 2, 3),
        (lambda x, y: x.rfloordiv(y), L(2), 7, 3),
    ]
)
def test_binary_arithmetic(con, func, left, right, expected):
    expr = func(left, right)
    result = con.execute(expr)
    assert result == expected


@pytest.mark.parametrize(
    ('value', 'expected'),
    [
        (L('foo_bar'), 'text'),
        (L(5), 'integer'),
        (ibis.NA, 'null'),

        # TODO(phillipc): should this really be double?
        (L(1.2345), 'numeric'),
        (
            L(
                datetime(
                    2015, 9, 1,
                    hour=14, minute=48, second=5, microsecond=359000
                )
            ),
            'timestamp without time zone'
        ),
        (L(date(2015, 9, 1)), 'date'),
    ]
)
def test_typeof(con, value, expected):
    assert con.execute(value.typeof()) == expected


@pytest.mark.parametrize(('value', 'expected'), [(0, None), (5.5, 5.5)])
def test_nullifzero(con, value, expected):
    assert con.execute(L(value).nullifzero()) == expected


@pytest.mark.parametrize(('value', 'expected'), [('foo_bar', 7), ('', 0)])
def test_string_length(con, value, expected):
    assert con.execute(L(value).length()) == expected


@pytest.mark.parametrize(
    ('op', 'expected'),
    [
        (methodcaller('left', 3), 'foo'),
        (methodcaller('right', 3), 'bar'),
        (methodcaller('substr', 0, 3), 'foo'),
        (methodcaller('substr', 4, 3), 'bar'),
        (methodcaller('substr', 1), 'oo_bar'),
    ]
)
def test_string_substring(con, op, expected):
    value = L('foo_bar')
    assert con.execute(op(value)) == expected


@pytest.mark.parametrize(
    ('op', 'expected'),
    [
        (methodcaller('lstrip'), 'foo   '),
        (methodcaller('rstrip'), '   foo'),
        (methodcaller('strip'), 'foo'),
    ]
)
def test_string_strip(con, op, expected):
    value = L('   foo   ')
    assert con.execute(op(value)) == expected


@pytest.mark.parametrize(
    ('op', 'expected'),
    [
        (methodcaller('lpad', 6, ' '), '   foo'),
        (methodcaller('rpad', 6, ' '), 'foo   '),
    ]
)
def test_string_pad(con, op, expected):
    value = L('foo')
    assert con.execute(op(value)) == expected


def test_string_reverse(con):
    assert con.execute(L('foo').reverse()) == 'oof'


def test_string_upper(con):
    assert con.execute(L('foo').upper()) == 'FOO'


def test_string_lower(con):
    assert con.execute(L('FOO').lower()) == 'foo'


@pytest.mark.parametrize(
    ('value', 'op', 'expected'),
    [
        (L('foobar'), methodcaller('contains', 'bar'), True),
        (L('foobar'), methodcaller('contains', 'foo'), True),
        (L('foobar'), methodcaller('contains', 'baz'), False),
        (L('100%'), methodcaller('contains', '%'), True),
        (L('a_b_c'), methodcaller('contains', '_'), True),
    ]
)
def test_string_contains(con, op, value, expected):
    assert con.execute(op(value)) == expected


@pytest.mark.parametrize(
    ('value', 'expected'),
    [
        ('foo bar foo', 'Foo Bar Foo'),
        ('foobar Foo', 'Foobar Foo'),
    ]
)
def test_capitalize(con, value, expected):
    assert con.execute(L(value).capitalize()) == expected


def test_repeat(con):
    expr = L('bar ').repeat(3)
    assert con.execute(expr) == 'bar bar bar '


def test_re_replace(con):
    expr = L('fudge|||chocolate||candy').re_replace('\\|{2,3}', ', ')
    assert con.execute(expr) == 'fudge, chocolate, candy'


def test_translate(con):
    expr = L('faab').translate('a', 'b')
    assert con.execute(expr) == 'fbbb'


@pytest.mark.parametrize(
    ('value', 'expected'),
    [
        (L('a'), 0),
        (L('b'), 1),
        (L('d'), -1),
        (L(None).cast(dt.string), 3),
    ]
)
def test_find_in_set(con, value, expected):
    assert con.execute(value.find_in_set(list('abc') + [None])) == expected


@pytest.mark.parametrize(
    ('expr', 'expected'),
    [
        (L(None).isnull(), True),
        (L(1).isnull(), False),
        (L(None).notnull(), False),
        (L(1).notnull(), True),
    ]
)
def test_isnull_notnull(con, expr, expected):
    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ('expr', 'expected'),
    [
        (L('foobar').find('bar'), 3),
        (L('foobar').find('baz'), -1),

        (L('foobar').like('%bar'), True),
        (L('foobar').like('foo%'), True),
        (L('foobar').like('%baz%'), False),

        (L('foobar').like(['%bar']), True),
        (L('foobar').like(['foo%']), True),
        (L('foobar').like(['%baz%']), False),

        (L('foobar').like(['%bar', 'foo%']), True),

        (L('foobarfoo').replace('foo', 'H'), 'HbarH'),
        (L('a').ascii_str(), ord('a'))
    ]
)
def test_string_functions(con, expr, expected):
    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ('expr', 'expected'),
    [
        (L(-5).abs(), 5),
        (L(5).abs(), 5),
        (ibis.least(L(5), L(10), L(1)), 1),
        (ibis.greatest(L(5), L(10), L(1)), 10),

        (L(5.5).round(), 6.0),
        (L(5.556).round(2), 5.56),
        (L(5.556).ceil(), 6.0),
        (L(5.556).floor(), 5.0),
        (L(5.556).exp(), math.exp(5.556)),
        (L(5.556).sign(), 1),
        (L(-5.556).sign(), -1),
        (L(0).sign(), 0),
        (L(5.556).sqrt(), math.sqrt(5.556)),
        (L(5.556).log(2), math.log(5.556, 2)),
        (L(5.556).ln(), math.log(5.556)),
        (L(5.556).log2(), math.log(5.556, 2)),
        (L(5.556).log10(), math.log10(5.556)),
    ]
)
def test_math_functions(con, expr, expected):
    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ('expr', 'expected'),
    [
        (L('abcd').re_search('[a-z]'), True),
        (L('abcd').re_search('[\d]+'), False),
        (L('1222').re_search('[\d]+'), True),
    ]
)
def test_regexp(con, expr, expected):
    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ('expr', 'expected'),
    [
        (L('abcd').re_extract('([a-z]+)', 0), 'abcd'),
        (L('abcd').re_extract('(ab)(cd)', 1), 'cd'),

        # valid group number but no match => empty string
        (L('abcd').re_extract('(\d)', 0), ''),

        # match but not a valid group number => NULL
        (L('abcd').re_extract('abcd', 3), None),
    ]
)
def test_regexp_extract(con, expr, expected):
    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ('expr', 'expected'),
    [
        (ibis.NA.fillna(5), 5),
        (L(5).fillna(10), 5),
        (L(5).nullif(5), None),
        (L(10).nullif(5), 10),
    ]
)
def test_fillna_nullif(con, expr, expected):
    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ('expr', 'expected'),
    [
        (ibis.coalesce(5, None, 4), 5),
        (ibis.coalesce(ibis.NA, 4, ibis.NA), 4),
        (ibis.coalesce(ibis.NA, ibis.NA, 3.14), 3.14),
    ]
)
def test_coalesce(con, expr, expected):
    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ('expr', 'expected'),
    [
        (ibis.coalesce(ibis.NA, ibis.NA), None),
        (ibis.coalesce(ibis.NA, ibis.NA, ibis.NA.cast('double')), None),
        (
            ibis.coalesce(
                ibis.NA.cast('int8'),
                ibis.NA.cast('int8'),
                ibis.NA.cast('int8'),
            ),
            None,
        ),
    ]
)
def test_coalesce_all_na(con, expr, expected):
    assert con.execute(expr) == expected


def test_numeric_builtins_work(alltypes, df):
    expr = alltypes.double_col.fillna(0)
    result = expr.execute()
    expected = df.double_col.fillna(0)
    expected.name = 'tmp'
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ('op', 'pandas_op'),
    [
        (
            lambda t: (t.double_col > 20).ifelse(10, -20),
            lambda df: pd.Series(
                np.where(df.double_col > 20, 10, -20),
                dtype='int8'
            ),
        ),
        (
            lambda t: (t.double_col > 20).ifelse(10, -20).abs(),
            lambda df: pd.Series(
                np.where(df.double_col > 20, 10, -20),
                dtype='int8'
            ).abs(),
        ),
    ]
)
def test_ifelse(alltypes, df, op, pandas_op):
    expr = op(alltypes)
    result = expr.execute()
    result.name = None
    expected = pandas_op(df)
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ('func', 'pandas_func'),
    [
        # tier and histogram
        (
            lambda d: d.bucket([0, 10, 25, 50, 100]),
            lambda s: pd.cut(
                s, [0, 10, 25, 50, 100], right=False, labels=False,
            )
        ),
        (
            lambda d: d.bucket([0, 10, 25, 50], include_over=True),
            lambda s: pd.cut(
                s, [0, 10, 25, 50, np.inf], right=False, labels=False
            )
        ),
        (
            lambda d: d.bucket([0, 10, 25, 50], close_extreme=False),
            lambda s: pd.cut(s, [0, 10, 25, 50], right=False, labels=False),
        ),
        (
            lambda d: d.bucket(
                [0, 10, 25, 50], closed='right', close_extreme=False
            ),
            lambda s: pd.cut(
                s, [0, 10, 25, 50],
                include_lowest=False,
                right=True,
                labels=False,
            )
        ),
        (
            lambda d: d.bucket([10, 25, 50, 100], include_under=True),
            lambda s: pd.cut(
                s, [0, 10, 25, 50, 100], right=False, labels=False
            ),
        ),
    ]
)
def test_bucket(alltypes, df, func, pandas_func):
    expr = func(alltypes.double_col)
    result = expr.execute()
    expected = pandas_func(df.double_col).astype('category')
    tm.assert_series_equal(result, expected, check_names=False)


def test_category_label(alltypes, df):
    t = alltypes
    d = t.double_col

    bins = [0, 10, 25, 50, 100]
    labels = ['a', 'b', 'c', 'd']
    bucket = d.bucket(bins)
    expr = bucket.label(labels)
    result = expr.execute().astype('category', ordered=True)
    result.name = 'double_col'

    expected = pd.cut(df.double_col, bins, labels=labels, right=False)

    tm.assert_series_equal(result, expected)


def test_union(alltypes):
    t = alltypes

    expr = (t.group_by('string_col')
            .aggregate(t.double_col.sum().name('foo'))
            .sort_by('string_col'))

    t1 = expr.limit(4)
    t2 = expr.limit(4, offset=4)
    t3 = expr.limit(8)

    result = t1.union(t2).execute()
    expected = t3.execute()
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    ('func', 'pandas_func'),
    [
        (
            lambda t, cond: t.bool_col.count(),
            lambda df, cond: df.bool_col.count(),
        ),
        (
            lambda t, cond: t.bool_col.any(),
            lambda df, cond: df.bool_col.any(),
        ),
        (
            lambda t, cond: t.bool_col.all(),
            lambda df, cond: df.bool_col.all(),
        ),
        (
            lambda t, cond: t.bool_col.notany(),
            lambda df, cond: ~df.bool_col.any(),
        ),
        (
            lambda t, cond: t.bool_col.notall(),
            lambda df, cond: ~df.bool_col.all(),
        ),

        (
            lambda t, cond: t.double_col.sum(),
            lambda df, cond: df.double_col.sum(),
        ),
        (
            lambda t, cond: t.double_col.mean(),
            lambda df, cond: df.double_col.mean(),
        ),
        (
            lambda t, cond: t.double_col.min(),
            lambda df, cond: df.double_col.min(),
        ),
        (
            lambda t, cond: t.double_col.max(),
            lambda df, cond: df.double_col.max(),
        ),
        (
            lambda t, cond: t.double_col.var(),
            lambda df, cond: df.double_col.var(),
        ),
        (
            lambda t, cond: t.double_col.std(),
            lambda df, cond: df.double_col.std(),
        ),
        (
            lambda t, cond: t.double_col.var(how='sample'),
            lambda df, cond: df.double_col.var(ddof=1),
        ),
        (
            lambda t, cond: t.double_col.std(how='pop'),
            lambda df, cond: df.double_col.std(ddof=0),
        ),
        (
            lambda t, cond: t.bool_col.count(where=cond),
            lambda df, cond: df.bool_col[cond].count(),
        ),
        (
            lambda t, cond: t.double_col.sum(where=cond),
            lambda df, cond: df.double_col[cond].sum(),
        ),
        (
            lambda t, cond: t.double_col.mean(where=cond),
            lambda df, cond: df.double_col[cond].mean(),
        ),
        (
            lambda t, cond: t.double_col.min(where=cond),
            lambda df, cond: df.double_col[cond].min(),
        ),
        (
            lambda t, cond: t.double_col.max(where=cond),
            lambda df, cond: df.double_col[cond].max(),
        ),
        (
            lambda t, cond: t.double_col.var(where=cond),
            lambda df, cond: df.double_col[cond].var(),
        ),
        (
            lambda t, cond: t.double_col.std(where=cond),
            lambda df, cond: df.double_col[cond].std(),
        ),
        (
            lambda t, cond: t.double_col.var(where=cond, how='sample'),
            lambda df, cond: df.double_col[cond].var(),
        ),
        (
            lambda t, cond: t.double_col.std(where=cond, how='pop'),
            lambda df, cond: df.double_col[cond].std(ddof=0),
        ),
    ]
)
def test_aggregations(alltypes, df, func, pandas_func):
    table = alltypes.limit(100)
    df = df.head(table.count().execute())

    cond = table.string_col.isin(['1', '7'])
    expr = func(table, cond)
    result = expr.execute()
    expected = pandas_func(df, cond.execute())

    np.testing.assert_allclose(result, expected)


def test_not_contains(alltypes, df):
    n = 100
    table = alltypes.limit(n)
    expr = table.string_col.notin(['1', '7'])
    result = expr.execute()
    expected = ~df.head(n).string_col.isin(['1', '7'])
    tm.assert_series_equal(result, expected, check_names=False)


def test_group_concat(alltypes, df):
    expr = alltypes.string_col.group_concat()
    result = expr.execute()
    expected = ','.join(df.string_col.dropna())
    assert result == expected


def test_distinct_aggregates(alltypes, df):
    expr = alltypes.limit(100).double_col.nunique()
    result = expr.execute()
    assert result == df.head(100).double_col.nunique()


def test_not_exists(alltypes, df):
    t = alltypes
    t2 = t.view()

    expr = t[~(t.string_col == t2.string_col).any()]
    result = expr.execute()

    left, right = df, t2.execute()
    expected = left[left.string_col != right.string_col]

    tm.assert_frame_equal(
        result, expected,
        check_index_type=False,
        check_dtype=False,
    )


def test_interactive_repr_shows_error(alltypes):
    # #591. Doing this in PostgreSQL because so many built-in functions are
    # not available

    expr = alltypes.double_col.approx_median()

    with config.option_context('interactive', True):
        result = repr(expr)

    assert 'no translation rule' in result.lower()


def test_subquery(alltypes, df):
    t = alltypes

    expr = (t.mutate(d=t.double_col.fillna(0))
            .limit(1000)
            .group_by('string_col')
            .size())
    result = expr.execute().sort_values('string_col').reset_index(drop=True)
    expected = df.assign(
        d=df.double_col.fillna(0)
    ).head(1000).groupby('string_col').string_col.count().reset_index(
        name='count'
    ).sort_values('string_col').reset_index(drop=True)
    tm.assert_frame_equal(
        result,
        expected,

        # Python 2 + pandas inferred type here is 'mixed' because of SQLAlchemy
        # string type subclasses
        check_column_type=sys.version_info.major >= 3
    )


@pytest.mark.parametrize('func', ['mean', 'sum', 'min', 'max'])
def test_simple_window(alltypes, func, df):
    t = alltypes
    f = getattr(t.double_col, func)
    df_f = getattr(df.double_col, func)
    result = t.projection([
        (t.double_col - f()).name('double_col')
    ]).execute().double_col
    expected = df.double_col - df_f()
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize('func', ['mean', 'sum', 'min', 'max'])
def test_rolling_window(alltypes, func, df):
    t = alltypes
    df = df[['double_col', 'timestamp_col']].sort_values(
        'timestamp_col'
    ).reset_index(drop=True)
    window = ibis.window(
        order_by=t.timestamp_col,
        preceding=6,
        following=0
    )
    f = getattr(t.double_col, func)
    df_f = getattr(df.double_col.rolling(7, min_periods=0), func)
    result = t.projection([
        f().over(window).name('double_col')
    ]).execute().double_col
    expected = df_f()
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize('func', ['mean', 'sum', 'min', 'max'])
def test_partitioned_window(alltypes, func, df):
    t = alltypes
    window = ibis.window(
        group_by=t.string_col,
        order_by=t.timestamp_col,
        preceding=6,
        following=0,
    )

    def roller(func):
        def rolled(df):
            torder = df.sort_values('timestamp_col')
            rolling = torder.double_col.rolling(7, min_periods=0)
            return getattr(rolling, func)()
        return rolled

    f = getattr(t.double_col, func)
    expr = f().over(window).name('double_col')
    result = t.projection([expr]).execute().double_col
    expected = df.groupby('string_col').apply(
        roller(func)
    ).reset_index(drop=True)
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize('func', ['sum', 'min', 'max'])
def test_cumulative_simple_window(alltypes, func, df):
    t = alltypes
    f = getattr(t.double_col, func)
    col = t.double_col - f().over(ibis.cumulative_window())
    expr = t.projection([col.name('double_col')])
    result = expr.execute().double_col
    expected = df.double_col - getattr(df.double_col, 'cum%s' % func)()
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize('func', ['sum', 'min', 'max'])
def test_cumulative_partitioned_window(alltypes, func, df):
    t = alltypes
    df = df.sort_values('string_col').reset_index(drop=True)
    window = ibis.cumulative_window(group_by=t.string_col)
    f = getattr(t.double_col, func)
    expr = t.projection([
        (t.double_col - f().over(window)).name('double_col')
    ])
    result = expr.execute().double_col
    expected = df.groupby(df.string_col).double_col.transform(
        lambda c: c - getattr(c, 'cum%s' % func)()
    )
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize('func', ['sum', 'min', 'max'])
def test_cumulative_ordered_window(alltypes, func, df):
    t = alltypes
    df = df.sort_values('timestamp_col').reset_index(drop=True)
    window = ibis.cumulative_window(order_by=t.timestamp_col)
    f = getattr(t.double_col, func)
    expr = t.projection([
        (t.double_col - f().over(window)).name('double_col')
    ])
    result = expr.execute().double_col
    expected = df.double_col - getattr(df.double_col, 'cum%s' % func)()
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize('func', ['sum', 'min', 'max'])
def test_cumulative_partitioned_ordered_window(alltypes, func, df):
    t = alltypes
    df = df.sort_values(['string_col', 'timestamp_col']).reset_index(drop=True)
    window = ibis.cumulative_window(
        order_by=t.timestamp_col, group_by=t.string_col
    )
    f = getattr(t.double_col, func)
    expr = t.projection([
        (t.double_col - f().over(window)).name('double_col')
    ])
    result = expr.execute().double_col
    expected = df.groupby(df.string_col).double_col.transform(
        lambda c: c - getattr(c, 'cum%s' % func)()
    )
    tm.assert_series_equal(result, expected)


def test_null_column(alltypes):
    t = alltypes
    nrows = t.count().execute()
    expr = t.mutate(na_column=ibis.NA).na_column
    result = expr.execute()
    tm.assert_series_equal(
        result,
        pd.Series([None] * nrows, name='na_column')
    )


def test_null_column_union(alltypes, df):
    t = alltypes
    s = alltypes[['double_col']].mutate(
        string_col=ibis.NA.cast('string'),
    )
    expr = t[['double_col', 'string_col']].union(s)
    result = expr.execute()
    nrows = t.count().execute()
    expected = pd.concat(
        [
            df[['double_col', 'string_col']],
            pd.concat(
                [
                    df[['double_col']],
                    pd.DataFrame({'string_col': [None] * nrows})
                ],
                axis=1,
            )
        ],
        axis=0,
        ignore_index=True
    )
    tm.assert_frame_equal(result, expected)


def test_window_with_arithmetic(alltypes, df):
    t = alltypes
    w = ibis.window(order_by=t.timestamp_col)
    expr = t.mutate(new_col=ibis.row_number().over(w) / 2)

    df = df[['timestamp_col']].sort_values('timestamp_col').reset_index(
        drop=True
    )
    expected = df.assign(new_col=[x / 2. for x in range(len(df))])
    result = expr['timestamp_col', 'new_col'].execute()
    tm.assert_frame_equal(result, expected)


def test_anonymous_aggregate(alltypes, df):
    t = alltypes
    expr = t[t.double_col > t.double_col.mean()]
    result = expr.execute()
    expected = df[df.double_col > df.double_col.mean()].reset_index(
        drop=True
    )
    tm.assert_frame_equal(result, expected)


@pytest.fixture
def array_types(con):
    return con.table('array_types')


def test_array_length(array_types):
    expr = array_types.projection([
        array_types.x.length().name('x_length'),
        array_types.y.length().name('y_length'),
        array_types.z.length().name('z_length'),
    ])
    result = expr.execute()
    expected = pd.DataFrame({
        'x_length': [3, 2, 2, 3, 3, 4],
        'y_length': [3, 2, 2, 3, 3, 4],
        'z_length': [3, 2, 2, 0, None, 4],
    })

    tm.assert_frame_equal(result, expected)


def test_array_schema(array_types):
    assert array_types.x.type() == dt.Array(dt.int64)
    assert array_types.y.type() == dt.Array(dt.string)
    assert array_types.z.type() == dt.Array(dt.double)


def test_array_collect(array_types):
    expr = array_types.group_by(
        array_types.grouper
    ).aggregate(collected=array_types.scalar_column.collect())
    result = expr.execute().sort_values('grouper').reset_index(drop=True)
    expected = pd.DataFrame({
        'grouper': list('abc'),
        'collected': [[1.0, 2.0, 3.0], [4.0, 5.0], [6.0]],
    })[['grouper', 'collected']]
    tm.assert_frame_equal(result, expected, check_column_type=False)


@pytest.mark.parametrize(
    ['start', 'stop'],
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),

        (None, 3),
        (None, None),
        (3, None),

        # negative slices are not supported
        pytest.mark.xfail(
            (-3, None),
            raises=ValueError,
            reason='Negative slicing not supported'
        ),
        pytest.mark.xfail(
            (None, -3),
            raises=ValueError,
            reason='Negative slicing not supported'
        ),
        pytest.mark.xfail(
            (-3, -1),
            raises=ValueError,
            reason='Negative slicing not supported'
        ),
    ]
)
def test_array_slice(array_types, start, stop):
    expr = array_types[array_types.y[start:stop].name('sliced')]
    result = expr.execute()
    expected = pd.DataFrame({
        'sliced': array_types.y.execute().map(lambda x: x[start:stop])
    })
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize('index', [1, 3, 4, 11])
def test_array_index(array_types, index):
    expr = array_types[array_types.y[index].name('indexed')]
    result = expr.execute()
    expected = pd.DataFrame({
        'indexed': array_types.y.execute().map(
            lambda x: x[index] if index < len(x) else None
        )
    })
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize('n', [1, 3, 4, 7, -2])
@pytest.mark.parametrize('mul', [lambda x, n: x * n, lambda x, n: n * x])
def test_array_repeat(array_types, n, mul):
    expr = array_types.projection([mul(array_types.x, n).name('repeated')])
    result = expr.execute()
    expected = pd.DataFrame({
        'repeated': array_types.x.execute().map(lambda x, n=n: mul(x, n))
    })
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize('catop', [lambda x, y: x + y, lambda x, y: y + x])
def test_array_concat(array_types, catop):
    t = array_types
    x, y = t.x.cast('array<string>').name('x'), t.y
    expr = t.projection([catop(x, y).name('catted')])
    result = expr.execute()
    tuples = t.projection([x, y]).execute().itertuples(index=False)
    expected = pd.DataFrame({'catted': [catop(i, j) for i, j in tuples]})
    tm.assert_frame_equal(result, expected)


def test_array_concat_mixed_types(array_types):
    with pytest.raises(TypeError):
        array_types.x + array_types.x.cast('array<double>')


@pytest.fixture
def t(con, guid):
    con.raw_sql(
        """
        CREATE TABLE "{}" (
          id SERIAL PRIMARY KEY,
          name TEXT
        )
        """.format(guid)
    )
    return con.table(guid)


@pytest.fixture
def s(con, t, guid, guid2):
    assert t.op().name == guid
    assert t.op().name != guid2

    con.raw_sql(
        """
        CREATE TABLE "{}" (
          id SERIAL PRIMARY KEY,
          left_t_id INTEGER REFERENCES "{}",
          cost DOUBLE PRECISION
        )
        """.format(guid2, guid)
    )
    return con.table(guid2)


@pytest.fixture
def trunc(con, guid):
    con.raw_sql(
        """
        CREATE TABLE "{}" (
          id SERIAL PRIMARY KEY,
          name TEXT
        )
        """.format(guid)
    )
    con.raw_sql(
        """INSERT INTO "{}" (name) VALUES ('a'), ('b'), ('c')""".format(
            guid
        )
    )
    return con.table(guid)


def test_semi_join(t, s):
    t_a, s_a = t.op().sqla_table.alias('t0'), s.op().sqla_table.alias('t1')
    expr = t.semi_join(s, t.id == s.id)
    result = expr.compile().compile(compile_kwargs=dict(literal_binds=True))
    base = sa.select([t_a.c.id, t_a.c.name]).where(
        sa.exists(sa.select([1]).where(t_a.c.id == s_a.c.id))
    )
    expected = sa.select([base.c.id, base.c.name])
    assert str(result) == str(expected)


def test_anti_join(t, s):
    t_a, s_a = t.op().sqla_table.alias('t0'), s.op().sqla_table.alias('t1')
    expr = t.anti_join(s, t.id == s.id)
    result = expr.compile().compile(compile_kwargs=dict(literal_binds=True))
    expected = sa.select([sa.column('id'), sa.column('name')]).select_from(
        sa.select([t_a.c.id, t_a.c.name]).where(
            ~sa.exists(sa.select([1]).where(t_a.c.id == s_a.c.id))
        )
    )
    assert str(result) == str(expected)


def test_create_table_from_expr(con, trunc, guid2):
    con.create_table(guid2, expr=trunc)
    t = con.table(guid2)
    assert list(t.name.execute()) == list('abc')


def test_truncate_table(con, trunc):
    assert list(trunc.name.execute()) == list('abc')
    con.truncate_table(trunc.op().name)
    assert not len(trunc.execute())


def test_head(con):
    t = con.table('functional_alltypes')
    result = t.head().execute()
    expected = t.limit(5).execute()
    tm.assert_frame_equal(result, expected)


def test_identical_to(con, df):
    # TODO: abstract this testing logic out into parameterized fixtures
    t = con.table('functional_alltypes')
    dt = df[['tinyint_col', 'double_col']]
    expr = t.tinyint_col.identical_to(t.double_col)
    result = expr.execute()
    expected = (dt.tinyint_col.isnull() & dt.double_col.isnull()) | (
        dt.tinyint_col == dt.double_col
    )
    expected.name = result.name
    tm.assert_series_equal(result, expected)


def test_rank(con):
    t = con.table('functional_alltypes')
    expr = t.double_col.rank()
    sqla_expr = expr.compile()
    result = str(sqla_expr.compile(compile_kwargs=dict(literal_binds=True)))
    expected = ("SELECT rank() OVER (ORDER BY t0.double_col ROWS BETWEEN "
                "UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) - 1 AS tmp \n"
                "FROM functional_alltypes AS t0")
    assert result == expected


def test_percent_rank(con):
    t = con.table('functional_alltypes')
    expr = t.double_col.percent_rank()
    sqla_expr = expr.compile()
    result = str(sqla_expr.compile(compile_kwargs=dict(literal_binds=True)))
    expected = ("SELECT percent_rank() OVER (ORDER BY t0.double_col ROWS "
                "BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) AS "
                "tmp \nFROM functional_alltypes AS t0")
    assert result == expected


def test_ntile(con):
    t = con.table('functional_alltypes')
    expr = t.double_col.ntile(7)
    sqla_expr = expr.compile()
    result = str(sqla_expr.compile(compile_kwargs=dict(literal_binds=True)))
    expected = ("SELECT ntile(7) OVER (ORDER BY t0.double_col ROWS BETWEEN "
                "UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING) - 1 AS tmp \n"
                "FROM functional_alltypes AS t0")
    assert result == expected


@pytest.mark.parametrize('op', [operator.invert, operator.neg])
def test_not_and_negate_bool(con, op, df):
    t = con.table('functional_alltypes').limit(10)
    expr = t.projection([op(t.bool_col).name('bool_col')])
    result = expr.execute().bool_col
    expected = op(df.head(10).bool_col)
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    'field',
    [
        'tinyint_col',
        'smallint_col',
        'int_col',
        'bigint_col',
        'float_col',
        'double_col',
        'year',
        'month',
    ]
)
def test_negate_non_boolean(con, field, df):
    t = con.table('functional_alltypes').limit(10)
    expr = t.projection([(-t[field]).name(field)])
    result = expr.execute()[field]
    expected = -df.head(10)[field]
    tm.assert_series_equal(result, expected)


def test_negate_boolean(con, df):
    t = con.table('functional_alltypes').limit(10)
    expr = t.projection([(-t.bool_col).name('bool_col')])
    result = expr.execute().bool_col
    expected = -df.head(10).bool_col
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ('attr', 'expected'),
    [
        (operator.methodcaller('year'), {2009, 2010}),
        (operator.methodcaller('month'), set(range(1, 13))),
        (operator.methodcaller('day'), set(range(1, 32)))
    ]
)
def test_date_extract_field(db, attr, expected):
    t = db.functional_alltypes
    expr = attr(t.timestamp_col.cast('date')).distinct()
    result = expr.execute().astype(int)
    assert set(result) == expected


@pytest.mark.parametrize(
    'op',
    list(map(
        operator.methodcaller, ['sum', 'mean', 'min', 'max', 'std', 'var']
    ))
)
def test_boolean_reduction(alltypes, op, df):
    result = op(alltypes.bool_col).execute()
    assert result == op(df.bool_col)


def test_boolean_summary(alltypes):
    expr = alltypes.bool_col.summary()
    result = expr.execute()
    expected = pd.DataFrame(
        [[7300, 0, 0, 1, 3650, 0.5, 2]],
        columns=[
            'count',
            'nulls',
            'min',
            'max',
            'sum',
            'mean',
            'approx_nunique',
        ]
    )

    type_conversions = {
        'count': 'int64',
        'nulls': 'int64',
        'min': 'bool',
        'max': 'bool',
        'sum': 'int64',
        'approx_nunique': 'int64'
    }
    for k, v in type_conversions.items():
        expected[k] = expected[k].astype(v)

    tm.assert_frame_equal(result, expected)


def test_timestamp_with_timezone(con):
    t = con.table('tzone')
    result = t.ts.execute()
    assert str(result.dtype.tz)


@pytest.fixture(
    params=[
        None,
        'UTC',
        'America/New_York',
        'America/Los_Angeles',
        'Europe/Paris',
        'Chile/Continental',
        'Asia/Tel_Aviv',
        'Asia/Tokyo',
        'Africa/Nairobi',
        'Australia/Sydney',
    ]
)
def tz(request):
    return request.param


@pytest.yield_fixture
def tzone_compute(con, guid, tz):
    schema = ibis.schema([
        ('ts', dt.timestamp(tz)),
        ('b', 'double'),
        ('c', 'string'),
    ])
    con.create_table(guid, schema=schema)
    t = con.table(guid)

    n = 10
    df = pd.DataFrame({
        'ts': pd.date_range('2017-04-01', periods=n, tz=tz).values,
        'b': np.arange(n).astype('float64'),
        'c': list(string.ascii_lowercase[:n]),
    })

    df.to_sql(
        guid,
        con.con,
        index=False,
        if_exists='append',
        dtype={
            'ts': sa.TIMESTAMP(timezone=True),
            'b': sa.FLOAT,
            'c': sa.TEXT,
        }
    )

    try:
        yield t
    finally:
        con.drop_table(guid)
        assert guid not in con.list_tables()


def test_ts_timezone_is_preserved(tzone_compute, tz):
    assert dt.Timestamp(tz).equals(tzone_compute.ts.type())


def test_timestamp_with_timezone_select(tzone_compute, tz):
    ts = tzone_compute.ts.execute()
    assert str(getattr(ts.dtype, 'tz', None)) == str(tz)


def test_timestamp_type_accepts_all_timezones(con):
    assert all(
        dt.Timestamp(row.name).timezone == row.name
        for row in con.con.execute(
            'SELECT name FROM pg_timezone_names'
        )
    )


@pytest.mark.parametrize(
    ('left', 'right', 'type'),
    [
        (L('2017-04-01'), date(2017, 4, 2), dt.date),
        (date(2017, 4, 2), L('2017-04-01'), dt.date),
        (
            L('2017-04-01 01:02:33'),
            datetime(2017, 4, 1, 1, 3, 34),
            dt.timestamp
        ),
        (
            datetime(2017, 4, 1, 1, 3, 34),
            L('2017-04-01 01:02:33'),
            dt.timestamp
        ),
    ]
)
@pytest.mark.parametrize(
    'op',
    [
        operator.eq,
        operator.ne,
        operator.lt,
        operator.le,
        operator.gt,
        operator.ge,
    ]
)
def test_string_temporal_compare(con, op, left, right, type):
    expr = op(left, right)
    result = con.execute(expr)
    left_raw = con.execute(L(left).cast(type))
    right_raw = con.execute(L(right).cast(type))
    expected = op(left_raw, right_raw)
    assert result == expected


@pytest.mark.parametrize(
    ('left', 'right'),
    [
        (L('2017-03-31').cast(dt.date), date(2017, 4, 2)),
        (date(2017, 3, 31), L('2017-04-02').cast(dt.date)),
        (
            L('2017-03-31 00:02:33').cast(dt.timestamp),
            datetime(2017, 4, 1, 1, 3, 34),
        ),
        (
            datetime(2017, 3, 31, 0, 2, 33),
            L('2017-04-01 01:03:34').cast(dt.timestamp),
        ),
    ]
)
@pytest.mark.parametrize(
    'op',
    [
        lambda left, right: ibis.timestamp('2017-04-01 00:02:34').between(
            left, right
        ),
        lambda left, right: ibis.timestamp('2017-04-01').cast(dt.date).between(
            left, right
        ),
    ]
)
def test_string_temporal_compare_between(con, op, left, right):
    expr = op(left, right)
    result = con.execute(expr)
    assert isinstance(result, (bool, np.bool_))
    assert result


def test_scalar_parameter(con):
    start = ibis.param(dt.date)
    end = ibis.param(dt.date)
    t = con.table('functional_alltypes')
    col = t.date_string_col.cast('date')
    expr = col.between(start, end)
    start_string, end_string = '2009-03-01', '2010-07-03'
    result = expr.execute(params={start: start_string, end: end_string})
    expected = col.between(start_string, end_string).execute()
    tm.assert_series_equal(result, expected)


def test_string_to_binary_cast(con):
    t = con.table('functional_alltypes').limit(10)
    expr = t.string_col.cast('binary')
    result = expr.execute()
    sql_string = (
        "SELECT decode(string_col, 'escape') AS tmp "
        "FROM functional_alltypes LIMIT 10"
    )
    raw_data = [row[0][0] for row in con.raw_sql(sql_string).fetchall()]
    expected = pd.Series(raw_data, name='tmp')
    tm.assert_series_equal(result, expected)


def test_string_to_binary_round_trip(con):
    t = con.table('functional_alltypes').limit(10)
    expr = t.string_col.cast('binary').cast('string')
    result = expr.execute()
    sql_string = (
        "SELECT encode(decode(string_col, 'escape'), 'escape') AS tmp "
        "FROM functional_alltypes LIMIT 10"
    )
    expected = pd.Series(
        [row[0][0] for row in con.raw_sql(sql_string).fetchall()],
        name='tmp'
    )
    tm.assert_series_equal(result, expected)



from six import StringIO
from datetime import date, datetime

import ibis.common as com
import ibis.util as util
import ibis.expr.types as ir
import ibis.expr.operations as ops
import ibis.sql.transforms as transforms

from ibis.clickhouse.identifiers import quote_identifier


def _cast(translator, expr):
    from ibis.clickhouse.client import ClickhouseDataType

    op = expr.op()
    arg, target = op.args
    arg_ = translator.translate(arg)
    type_ = str(ClickhouseDataType.from_ibis(target, nullable=False))

    return 'CAST({0!s} AS {1!s})'.format(arg_, type_)


def _between(translator, expr):
    op = expr.op()
    arg_, lower_, upper_ = map(translator.translate, op.args)
    return '{0!s} BETWEEN {1!s} AND {2!s}'.format(arg_, lower_, upper_)


def _negate(translator, expr):
    arg = expr.op().args[0]
    if isinstance(expr, ir.BooleanValue):
        arg_ = translator.translate(arg)
        return 'NOT {0!s}'.format(arg_)
    else:
        arg_ = _parenthesize(translator, arg)
        return '-{0!s}'.format(arg_)


def _not(translator, expr):
    return 'NOT {}'.format(*map(translator.translate, expr.op().args))


def _parenthesize(translator, expr):
    op = expr.op()
    op_klass = type(op)

    # function calls don't need parens
    what_ = translator.translate(expr)
    if (op_klass in _binary_infix_ops) or (op_klass in _unary_ops):
        return '({0!s})'.format(what_)
    else:
        return what_


def unary(func_name):
    return fixed_arity(func_name, 1)


def fixed_arity(func_name, arity):
    def formatter(translator, expr):
        op = expr.op()
        arg_count = len(op.args)
        if arity != arg_count:
            msg = 'Incorrect number of args {0} instead of {1}'
            raise com.UnsupportedOperationError(msg.format(arg_count, arity))
        return _call(translator, func_name, *op.args)
    return formatter


def agg(func):
    def formatter(translator, expr):
        return _aggregate(translator, func, *expr.op().args)
    return formatter


def agg_variance_like(func):
    variants = {'sample': '{0}Samp'.format(func),
                'pop': '{0}Pop'.format(func)}

    def formatter(translator, expr):
        arg, how, where = expr.op().args
        return _aggregate(translator, variants[how], arg, where)

    return formatter


def binary_infix_op(infix_sym):
    def formatter(translator, expr):
        op = expr.op()

        left, right = op.args
        left_ = _parenthesize(translator, left)
        right_ = _parenthesize(translator, right)

        return '{0!s} {1!s} {2!s}'.format(left_, infix_sym, right_)
    return formatter


def _call(translator, func, *args):
    args_ = ', '.join(map(translator.translate, args))
    return '{0!s}({1!s})'.format(func, args_)


def _aggregate(translator, func, arg, where=None):
    if where is not None:
        return _call(translator, func + 'If', arg, where)
    else:
        return _call(translator, func, arg)


def _xor(translator, expr):
    op = expr.op()
    left_ = _parenthesize(translator, op.left)
    right_ = _parenthesize(translator, op.right)
    return 'xor({0}, {1})'.format(left_, right_)


def _name_expr(formatted_expr, quoted_name):
    return '{0!s} AS {1!s}'.format(formatted_expr, quoted_name)


def varargs(func_name):
    def varargs_formatter(translator, expr):
        op = expr.op()
        return _call(translator, func_name, *op.args)
    return varargs_formatter


def _arbitrary(translator, expr):
    arg, how, where = expr.op().args
    functions = {'first': 'any',
                 'last': 'anyLast',
                 'heavy': 'anyHeavy'}
    return _aggregate(translator, functions[how], arg, where=where)


def _substring(translator, expr):
    # arg_ is the formatted notation
    op = expr.op()
    arg, start, length = op.args
    arg_, start_ = translator.translate(arg), translator.translate(start)

    # Clickhouse is 1-indexed
    if length is None or isinstance(length.op(), ir.Literal):
        if length is not None:
            length_ = length.op().value
            return 'substring({0}, {1} + 1, {2})'.format(arg_, start_, length_)
        else:
            return 'substring({0}, {1} + 1)'.format(arg_, start_)
    else:
        length_ = translator.translate(length)
        return 'substring({0}, {1} + 1, {2})'.format(arg_, start_, length_)


def _string_find(translator, expr):
    op = expr.op()
    arg, substr, start, _ = op.args
    if start is not None:
        raise com.UnsupportedOperationError(
            "String find doesn't support start argument"
        )

    return _call(translator, 'position', arg, substr) + ' - 1'


def _regex_extract(translator, expr):
    op = expr.op()
    arg, pattern, index = op.args
    arg_, pattern_ = translator.translate(arg), translator.translate(pattern)

    if index is not None:
        index_ = translator.translate(index)
        return 'extractAll({0}, {1})[{2} + 1]'.format(arg_, pattern_, index_)

    return 'extractAll({0}, {1})'.format(arg_, pattern_)


def _parse_url(translator, expr):
    op = expr.op()
    arg, extract, key = op.args

    if extract == 'HOST':
        return _call(translator, 'domain', arg)
    elif extract == 'PROTOCOL':
        return _call(translator, 'protocol', arg)
    elif extract == 'PATH':
        return _call(translator, 'path', arg)
    elif extract == 'QUERY':
        if key is not None:
            return _call(translator, 'extractURLParameter', arg, key)
        else:
            return _call(translator, 'queryString', arg)
    else:
        raise com.UnsupportedOperationError(
            'Parse url with extract {0} is not supported'.format(extract)
        )


def _index_of(translator, expr):
    op = expr.op()

    arg, arr = op.args
    arg_formatted = translator.translate(arg)
    arr_formatted = ','.join(map(translator.translate, arr))
    return "indexOf([{0}], {1}) - 1".format(arr_formatted, arg_formatted)


def _sign(translator, expr):
    """Workaround for missing sign function"""
    op = expr.op()
    arg, = op.args
    arg_ = translator.translate(arg)
    return 'intDivOrZero({0}, abs({0}))'.format(arg_)


def _round(translator, expr):
    op = expr.op()
    arg, digits = op.args

    if digits is not None:
        return _call(translator, 'round', arg, digits)
    else:
        return _call(translator, 'round', arg)


def _hash(translator, expr):
    op = expr.op()
    arg, how = op.args

    algorithms = {'MD5', 'halfMD5',
                  'SHA1', 'SHA224', 'SHA256',
                  'intHash32', 'intHash64',
                  'cityHash64',
                  'sipHash64', 'sipHash128'}

    if how not in algorithms:
        raise com.UnsupportedOperationError(
            'Unsupported hash algorithm {0}'.format(how)
        )

    return _call(translator, how, arg)


def _log(translator, expr):
    op = expr.op()
    arg, base = op.args

    if base is None:
        func = 'log'
    elif base._arg.value == 2:
        func = 'log2'
    elif base._arg.value == 10:
        func = 'log10'
    else:
        raise ValueError('Base {} for logarithm not supported!'.format(base))

    return _call(translator, func, arg)


def _value_list(translator, expr):
    op = expr.op()
    values_ = map(translator.translate, op.values)
    return '({0})'.format(', '.join(values_))


def _interval_format(translator, expr):
    if expr.unit in {'ms', 'us', 'ns'}:
        raise com.UnsupportedOperationError(
            "Clickhouse doesn't support subsecond interval resolutions")

    return 'INTERVAL {} {}'.format(expr.op().value, expr.resolution.upper())


def _interval_from_integer(translator, expr):
    op = expr.op()
    arg, unit = op.args

    if expr.unit in {'ms', 'us', 'ns'}:
        raise com.UnsupportedOperationError(
            "Clickhouse doesn't support subsecond interval resolutions")

    arg_ = translator.translate(arg)
    return 'INTERVAL {} {}'.format(arg_, expr.resolution.upper())


def literal(translator, expr):
    value = expr.op().value
    if isinstance(expr, ir.BooleanValue):
        return '1' if value else '0'
    elif isinstance(expr, ir.StringValue):
        return "'{0!s}'".format(value.replace("'", "\\'"))
    elif isinstance(expr, ir.NumericValue):
        return repr(value)
    elif isinstance(expr, ir.IntervalValue):
        return _interval_format(translator, expr)
    elif isinstance(expr, ir.TimestampValue):
        if isinstance(value, datetime):
            if value.microsecond != 0:
                msg = 'Unsupported subsecond accuracy {}'
                raise ValueError(msg.format(value))
            value = value.strftime('%Y-%m-%d %H:%M:%S')
        return "toDateTime('{0!s}')".format(value)
    elif isinstance(expr, ir.DateValue):
        if isinstance(value, date):
            value = value.strftime('%Y-%m-%d')
        return "toDate('{0!s}')".format(value)
    elif isinstance(expr, ir.ArrayValue):
        return str(list(value))
    else:
        raise NotImplementedError(type(expr))


class CaseFormatter(object):

    def __init__(self, translator, base, cases, results, default):
        self.translator = translator
        self.base = base
        self.cases = cases
        self.results = results
        self.default = default

        # HACK
        self.indent = 2
        self.multiline = len(cases) > 1
        self.buf = StringIO()

    def _trans(self, expr):
        return self.translator.translate(expr)

    def get_result(self):
        self.buf.seek(0)

        self.buf.write('CASE')
        if self.base is not None:
            base_str = self._trans(self.base)
            self.buf.write(' {0}'.format(base_str))

        for case, result in zip(self.cases, self.results):
            self._next_case()
            case_str = self._trans(case)
            result_str = self._trans(result)
            self.buf.write('WHEN {0} THEN {1}'.format(case_str, result_str))

        if self.default is not None:
            self._next_case()
            default_str = self._trans(self.default)
            self.buf.write('ELSE {0}'.format(default_str))

        if self.multiline:
            self.buf.write('\nEND')
        else:
            self.buf.write(' END')

        return self.buf.getvalue()

    def _next_case(self):
        if self.multiline:
            self.buf.write('\n{0}'.format(' ' * self.indent))
        else:
            self.buf.write(' ')


def _simple_case(translator, expr):
    op = expr.op()
    formatter = CaseFormatter(translator, op.base, op.cases, op.results,
                              op.default)
    return formatter.get_result()


def _searched_case(translator, expr):
    op = expr.op()
    formatter = CaseFormatter(translator, None, op.cases, op.results,
                              op.default)
    return formatter.get_result()


def _table_array_view(translator, expr):
    ctx = translator.context
    table = expr.op().table
    query = ctx.get_compiled_expr(table)
    return '(\n{0}\n)'.format(util.indent(query, ctx.indent))


def _timestamp_from_unix(translator, expr):
    op = expr.op()
    arg, unit = op.args

    if unit in {'ms', 'us', 'ns'}:
        raise ValueError('`{}` unit is not supported!'.format(unit))

    return _call(translator, 'toDateTime', arg)


def _truncate(translator, expr):
    op = expr.op()
    arg, unit = op.args

    converters = {
        'Y': 'toStartOfYear',
        'M': 'toStartOfMonth',
        'W': 'toMonday',
        'D': 'toDate',
        'h': 'toStartOfHour',
        'm': 'toStartOfMinute',
        's': 'toDateTime'
    }

    try:
        converter = converters[unit]
    except KeyError:
        raise com.UnsupportedOperationError(
            'Unsupported truncate unit {}'.format(unit)
        )

    return _call(translator, converter, arg)


def _exists_subquery(translator, expr):
    op = expr.op()
    ctx = translator.context

    dummy = ir.literal(1).name(ir.unnamed)

    filtered = op.foreign_table.filter(op.predicates)
    expr = filtered.projection([dummy])

    subquery = ctx.get_compiled_expr(expr)

    if isinstance(op, transforms.ExistsSubquery):
        key = 'EXISTS'
    elif isinstance(op, transforms.NotExistsSubquery):
        key = 'NOT EXISTS'
    else:
        raise NotImplementedError

    return '{0} (\n{1}\n)'.format(key, util.indent(subquery, ctx.indent))


def _table_column(translator, expr):
    op = expr.op()
    field_name = op.name
    quoted_name = quote_identifier(field_name, force=True)
    table = op.table
    ctx = translator.context

    # If the column does not originate from the table set in the current SELECT
    # context, we should format as a subquery
    if translator.permit_subquery and ctx.is_foreign_expr(table):
        proj_expr = table.projection([field_name]).to_array()
        return _table_array_view(translator, proj_expr)

    # TODO(kszucs): table aliasing is partially supported
    # if ctx.need_aliases():
    #     alias = ctx.get_ref(table)
    #     if alias is not None:
    #         quoted_name = '{0}.{1}'.format(alias, quoted_name)

    return quoted_name


def _string_split(translator, expr):
    value, sep = expr.op().args
    return 'splitByString({}, {})'.format(
        translator.translate(sep),
        translator.translate(value)
    )


def _string_join(translator, expr):
    sep, elements = expr.op().args
    assert isinstance(elements.op(), ir.ValueList), \
        'elements must be a ValueList, got {}'.format(type(elements.op()))
    return 'arrayStringConcat([{}], {})'.format(
        ', '.join(map(translator.translate, elements)),
        translator.translate(sep),
    )


def _string_repeat(translator, expr):
    value, times = expr.op().args
    result = 'arrayStringConcat(arrayMap(x -> {}, range({})))'.format(
        translator.translate(value), translator.translate(times)
    )
    return result


def _string_like(translator, expr):
    value, pattern = expr.op().args[:2]
    return '{} LIKE {}'.format(
        translator.translate(value), translator.translate(pattern)
    )


# TODO: clickhouse uses different string functions
#       for ascii and utf-8 encodings,

_binary_infix_ops = {
    # Binary operations
    ops.Add: binary_infix_op('+'),
    ops.Subtract: binary_infix_op('-'),
    ops.Multiply: binary_infix_op('*'),
    ops.Divide: binary_infix_op('/'),
    ops.Power: fixed_arity('pow', 2),
    ops.Modulus: binary_infix_op('%'),

    # Comparisons
    ops.Equals: binary_infix_op('='),
    ops.NotEquals: binary_infix_op('!='),
    ops.GreaterEqual: binary_infix_op('>='),
    ops.Greater: binary_infix_op('>'),
    ops.LessEqual: binary_infix_op('<='),
    ops.Less: binary_infix_op('<'),

    # Boolean comparisons
    ops.And: binary_infix_op('AND'),
    ops.Or: binary_infix_op('OR'),
    ops.Xor: _xor,
}

_unary_ops = {
    ops.Negate: _negate,
    ops.Not: _not
}


_operation_registry = {
    # Unary operations
    ops.TypeOf: unary('toTypeName'),

    ops.IsNan: unary('isNaN'),
    ops.IsInf: unary('isInfinite'),

    ops.Abs: unary('abs'),
    ops.Ceil: unary('ceil'),
    ops.Floor: unary('floor'),
    ops.Exp: unary('exp'),
    ops.Round: _round,

    ops.Sign: _sign,
    ops.Sqrt: unary('sqrt'),

    ops.Hash: _hash,

    ops.Log: _log,
    ops.Ln: unary('log'),
    ops.Log2: unary('log2'),
    ops.Log10: unary('log10'),

    # Unary aggregates
    ops.CMSMedian: agg('median'),
    # TODO: there is also a `uniq` function which is the
    #       recommended way to approximate cardinality
    ops.HLLCardinality: agg('uniqHLL12'),
    ops.Mean: agg('avg'),
    ops.Sum: agg('sum'),
    ops.Max: agg('max'),
    ops.Min: agg('min'),

    ops.StandardDev: agg_variance_like('stddev'),
    ops.Variance: agg_variance_like('var'),

    # ops.GroupConcat: fixed_arity('group_concat', 2),

    ops.Count: agg('count'),
    ops.CountDistinct: agg('uniq'),
    ops.Arbitrary: _arbitrary,

    # string operations
    ops.StringLength: unary('length'),
    ops.Lowercase: unary('lower'),
    ops.Uppercase: unary('upper'),
    ops.Reverse: unary('reverse'),
    ops.Substring: _substring,
    ops.StringFind: _string_find,
    ops.FindInSet: _index_of,
    ops.StringReplace: fixed_arity('replaceAll', 3),
    ops.StringJoin: _string_join,
    ops.StringSplit: _string_split,
    ops.StringSQLLike: _string_like,
    ops.Repeat: _string_repeat,

    ops.RegexSearch: fixed_arity('match', 2),
    # TODO: extractAll(haystack, pattern)[index + 1]
    ops.RegexExtract: _regex_extract,
    ops.RegexReplace: fixed_arity('replaceRegexpAll', 3),
    ops.ParseURL: _parse_url,

    # Temporal operations
    ops.Date: unary('toDate'),
    ops.DateTruncate: _truncate,

    ops.TimestampNow: lambda *args: 'now()',
    ops.TimestampTruncate: _truncate,

    ops.TimeTruncate: _truncate,

    ops.IntervalFromInteger: _interval_from_integer,

    ops.ExtractYear: unary('toYear'),
    ops.ExtractMonth: unary('toMonth'),
    ops.ExtractDay: unary('toDayOfMonth'),
    ops.ExtractHour: unary('toHour'),
    ops.ExtractMinute: unary('toMinute'),
    ops.ExtractSecond: unary('toSecond'),

    # Other operations
    ops.E: lambda *args: 'e()',

    ir.Literal: literal,
    ir.ValueList: _value_list,

    ops.Cast: _cast,

    # for more than 2 args this should be arrayGreatest|Least(array([]))
    # because clickhouse's greatest and least doesn't support varargs
    ops.Greatest: varargs('greatest'),
    ops.Least: varargs('least'),

    ops.Where: fixed_arity('if', 3),

    ops.Between: _between,
    ops.Contains: binary_infix_op('IN'),
    ops.NotContains: binary_infix_op('NOT IN'),

    ops.SimpleCase: _simple_case,
    ops.SearchedCase: _searched_case,

    ops.TableColumn: _table_column,
    ops.TableArrayView: _table_array_view,

    ops.DateAdd: binary_infix_op('+'),
    ops.DateSub: binary_infix_op('-'),
    ops.DateDiff: binary_infix_op('-'),
    ops.TimestampAdd: binary_infix_op('+'),
    ops.TimestampSub: binary_infix_op('-'),
    ops.TimestampDiff: binary_infix_op('-'),
    ops.TimestampFromUNIX: _timestamp_from_unix,

    transforms.ExistsSubquery: _exists_subquery,
    transforms.NotExistsSubquery: _exists_subquery,

    ops.ArrayLength: unary('length'),
}


def raise_error(translator, expr, *args):
    msg = "Clickhouse backend doesn't support {0} operation!"
    op = expr.op()
    raise com.UnsupportedOperationError(msg.format(type(op)))


def _null_literal(translator, expr):
    return 'Null'


def _null_if_zero(translator, expr):
    op = expr.op()
    arg = op.args[0]
    arg_ = translator.translate(arg)
    return 'nullIf({0}, 0)'.format(arg_)


def _zero_if_null(translator, expr):
    op = expr.op()
    arg = op.args[0]
    arg_ = translator.translate(arg)
    return 'ifNull({0}, 0)'.format(arg_)


_undocumented_operations = {
    ir.NullLiteral: _null_literal,  # undocumented
    ops.IsNull: unary('isNull'),
    ops.NotNull: unary('isNotNull'),
    ops.IfNull: fixed_arity('ifNull', 2),
    ops.NullIf: fixed_arity('nullIf', 2),
    ops.Coalesce: varargs('coalesce'),
    ops.NullIfZero: _null_if_zero,
    ops.ZeroIfNull: _zero_if_null
}


_unsupported_ops = [
    ops.WindowOp,
    ops.DecimalPrecision,
    ops.DecimalScale,
    ops.BaseConvert,
    ops.CumulativeSum,
    ops.CumulativeMin,
    ops.CumulativeMax,
    ops.CumulativeMean,
    ops.CumulativeAny,
    ops.CumulativeAll,
    ops.IdenticalTo,

    ops.RowNumber,
    ops.DenseRank,
    ops.MinRank,
    ops.PercentRank,

    ops.FirstValue,
    ops.LastValue,
    ops.NthValue,
    ops.Lag,
    ops.Lead,
    ops.NTile
]
_unsupported_ops = {k: raise_error for k in _unsupported_ops}


_operation_registry.update(_undocumented_operations)
_operation_registry.update(_unsupported_ops)
_operation_registry.update(_unary_ops)
_operation_registry.update(_binary_infix_ops)



import pytest
import datetime
import operator

from ibis.compat import PY2
import ibis
import ibis.expr.api as api
import ibis.expr.types as ir
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops


@pytest.mark.parametrize('timedelta', [
    api.timedelta,
    api.year,
    api.quarter,
    api.month,
    api.week,
    api.day,
    api.hour,
    api.minute,
    api.second,
    api.millisecond,
    api.microsecond,
    api.nanosecond
])
def test_timedelta_deprecated(timedelta):
    with pytest.deprecated_call():
        timedelta(1)


def test_temporal_literals():
    date = ibis.literal('2015-01-01', 'date')
    assert isinstance(date, ir.DateScalar)

    timestamp = ibis.literal('2017-01-01 12:00:33', 'timestamp')
    assert isinstance(timestamp, ir.TimestampScalar)


@pytest.mark.parametrize(('interval', 'unit', 'expected'), [
    (api.month(3), 'Q', api.quarter(1)),
    (api.month(12), 'Y', api.year(1)),
    (api.quarter(8), 'Y', api.year(2)),
    (api.day(14), 'W', api.week(2)),
    (api.minute(240), 'h', api.hour(4)),
    (api.second(360), 'm', api.minute(6)),
    (api.second(3 * 86400), 'D', api.day(3)),
    (api.millisecond(5000), 's', api.second(5)),
    (api.microsecond(5000000), 's', api.second(5)),
    (api.nanosecond(5000000000), 's', api.second(5)),
])
def test_upconvert(interval, unit, expected):
    result = interval.to_unit(unit)

    assert isinstance(result, ir.IntervalScalar)
    assert result.type().unit == expected.type().unit


@pytest.mark.parametrize('target', [
    'Y', 'Q', 'M'
])
@pytest.mark.parametrize('delta', [
    api.week(),
    api.day(),
    api.hour(),
    api.minute(),
    api.second(),
    api.millisecond(),
    api.microsecond(),
    api.nanosecond()
])
def test_cannot_upconvert(delta, target):
    with pytest.raises(ValueError):
        delta.to_unit(target)


@pytest.mark.parametrize('expr', [
    api.day(2) * 2,
    api.day(2) * (-2),
    3 * api.day(2),
    (-3) * api.day(2)
])
def test_multiply(expr):
    assert isinstance(expr, ir.IntervalScalar)
    assert expr.unit == 'D'


@pytest.mark.parametrize(('case', 'expected'), [
    (api.second(2).to_unit('s'), api.second(2)),
    (api.second(2).to_unit('ms'), api.millisecond(2 * 1000)),
    (api.second(2).to_unit('us'), api.microsecond(2 * 1000000)),
    (api.second(2).to_unit('ns'), api.nanosecond(2 * 1000000000)),

    (api.millisecond(2).to_unit('ms'), api.millisecond(2)),
    (api.millisecond(2).to_unit('us'), api.microsecond(2 * 1000)),
    (api.millisecond(2).to_unit('ns'), api.nanosecond(2 * 1000000)),

    (api.microsecond(2).to_unit('us'), api.microsecond(2)),
    (api.microsecond(2).to_unit('ns'), api.nanosecond(2 * 1000)),

    (api.nanosecond(2).to_unit('ns'), api.nanosecond(2)),
])
def test_downconvert_second_parts(case, expected):
    assert isinstance(case, ir.IntervalScalar)
    assert isinstance(expected, ir.IntervalScalar)
    assert case.unit == expected.unit


@pytest.mark.parametrize(('case', 'expected'), [
    (api.hour(2).to_unit('h'), api.hour(2)),
    (api.hour(2).to_unit('m'), api.minute(2 * 60)),
    (api.hour(2).to_unit('s'), api.second(2 * 3600)),
    (api.hour(2).to_unit('ms'), api.millisecond(2 * 3600000)),
    (api.hour(2).to_unit('us'), api.microsecond(2 * 3600000000)),
    (api.hour(2).to_unit('ns'), api.nanosecond(2 * 3600000000000))
])
def test_downconvert_hours(case, expected):
    assert isinstance(case, ir.IntervalScalar)
    assert isinstance(expected, ir.IntervalScalar)
    assert case.unit == expected.unit


@pytest.mark.parametrize(('case', 'expected'), [
    (api.week(2).to_unit('D'), api.day(2 * 7)),
    (api.week(2).to_unit('h'), api.hour(2 * 7 * 24)),

    (api.day(2).to_unit('D'), api.day(2)),
    (api.day(2).to_unit('h'), api.hour(2 * 24)),
    (api.day(2).to_unit('m'), api.minute(2 * 1440)),
    (api.day(2).to_unit('s'), api.second(2 * 86400)),
    (api.day(2).to_unit('ms'), api.millisecond(2 * 86400000)),
    (api.day(2).to_unit('us'), api.microsecond(2 * 86400000000)),
    (api.day(2).to_unit('ns'), api.nanosecond(2 * 86400000000000)),
])
def test_downconvert_day(case, expected):
    assert isinstance(case, ir.IntervalScalar)
    assert isinstance(expected, ir.IntervalScalar)
    assert case.unit == expected.unit


@pytest.mark.parametrize(('a', 'b', 'unit'), [
    (api.day(), api.day(3), 'D'),
    (api.second(), api.hour(10), 's'),
    (api.hour(3), api.day(2), 'h')
])
def test_combine_with_different_kinds(a, b, unit):
    assert (a + b).type().unit == unit


@pytest.mark.parametrize(('case', 'expected'), [
    (api.interval(quarters=2), api.quarter(2)),
    (api.interval(weeks=2), api.week(2)),
    (api.interval(days=3), api.day(3)),
    (api.interval(hours=4), api.hour(4)),
    (api.interval(minutes=5), api.minute(5)),
    (api.interval(seconds=6), api.second(6)),
    (api.interval(milliseconds=7), api.millisecond(7)),
    (api.interval(microseconds=8), api.microsecond(8)),
    (api.interval(nanoseconds=9), api.nanosecond(9)),
])
def test_timedelta_generic_api(case, expected):
    assert case.equals(expected)


def test_interval_time_expr(table):
    c = table.k
    x = api.timedelta(hours=1)

    expr = x + c
    assert isinstance(expr, ir.TimeColumn)
    assert isinstance(expr.op(), ops.TimeAdd)

    # test radd
    expr = c + x
    assert isinstance(expr, ir.TimeColumn)
    assert isinstance(expr.op(), ops.TimeAdd)


def test_interval_date_expr(table):
    c = table.j
    x = api.timedelta(days=1)

    expr = x + c
    assert isinstance(expr, ir.DateColumn)
    assert isinstance(expr.op(), ops.DateAdd)

    # test radd
    expr = c + x
    assert isinstance(expr, ir.DateColumn)
    assert isinstance(expr.op(), ops.DateAdd)


def test_interval_timestamp_expr(table):
    c = table.i
    x = api.timedelta(seconds=1)

    expr = x + c
    assert isinstance(expr, ir.TimestampColumn)
    assert isinstance(expr.op(), ops.TimestampAdd)

    # test radd
    expr = c + x
    assert isinstance(expr, ir.TimestampColumn)
    assert isinstance(expr.op(), ops.TimestampAdd)


@pytest.mark.xfail(raises=AssertionError, reason='NYI')
def test_compound_offset():
    # These are not yet allowed (e.g. 1 month + 1 hour)
    assert False


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_offset_months():
    assert False


@pytest.mark.parametrize('literal', [
    api.interval(3600),
    api.interval(datetime.timedelta(days=3)),
    api.interval(years=1),
    api.interval(quarters=3),
    api.interval(months=2),
    api.interval(weeks=3),
    api.interval(days=-1),
    api.interval(hours=-3),
    api.interval(minutes=5),
    api.interval(seconds=-8)
])
def test_interval(literal):
    assert isinstance(literal, ir.IntervalScalar)


@pytest.mark.parametrize(('expr', 'expected'), [
    (api.interval(weeks=3), "Literal[interval<int8>(unit='W')]\n  3"),
    (api.interval(months=3), "Literal[interval<int8>(unit='M')]\n  3"),
    (api.interval(seconds=-10), "Literal[interval<int8>(unit='s')]\n  -10")
])
def test_interval_repr(expr, expected):
    assert repr(expr) == expected


def test_timestamp_arithmetics():
    ts1 = api.timestamp(datetime.datetime.now())
    ts2 = api.timestamp(datetime.datetime.today())

    i1 = api.interval(minutes=30)

    # TODO: raise for unsupported operations too
    for expr in [ts2 - ts1, ts1 - ts2]:
        assert isinstance(expr, ir.IntervalScalar)
        assert isinstance(expr.op(), ops.TimestampDiff)
        assert expr.type() == dt.Interval('s', dt.int32)

    for expr in [ts1 - i1, ts2 - i1]:
        assert isinstance(expr, ir.TimestampScalar)
        assert isinstance(expr.op(), ops.TimestampSub)

    for expr in [ts1 + i1, ts2 + i1]:
        assert isinstance(expr, ir.TimestampScalar)
        assert isinstance(expr.op(), ops.TimestampAdd)


def test_date_arithmetics():
    d1 = api.date('2015-01-02')
    d2 = api.date('2017-01-01')
    i1 = api.interval(weeks=3)

    for expr in [d1 - d2, d2 - d1]:
        assert isinstance(expr, ir.IntervalScalar)
        assert isinstance(expr.op(), ops.DateDiff)
        assert expr.type() == dt.Interval('D', dt.int32)

    for expr in [d1 - i1, d2 - i1]:
        assert isinstance(expr, ir.DateScalar)
        assert isinstance(expr.op(), ops.DateSub)

    for expr in [d1 + i1, d2 + i1]:
        assert isinstance(expr, ir.DateScalar)
        assert isinstance(expr.op(), ops.DateAdd)


@pytest.mark.skipif(PY2, reason='time support is not enabled on python 2')
def test_time_arithmetics():
    t1 = api.time('18:00')
    t2 = api.time('19:12')
    i1 = api.interval(minutes=3)

    for expr in [t1 - t2, t2 - t1]:
        assert isinstance(expr, ir.IntervalScalar)
        assert isinstance(expr.op(), ops.TimeDiff)
        assert expr.type() == dt.Interval('s', dt.int32)

    for expr in [t1 - i1, t2 - i1]:
        assert isinstance(expr, ir.TimeScalar)
        assert isinstance(expr.op(), ops.TimeSub)

    for expr in [t1 + i1, t2 + i1]:
        assert isinstance(expr, ir.TimeScalar)
        assert isinstance(expr.op(), ops.TimeAdd)


def test_invalid_date_arithmetics():
    d1 = api.date('2015-01-02')
    i1 = api.interval(seconds=300)
    i2 = api.interval(minutes=15)
    i3 = api.interval(hours=1)

    for i in [i1, i2, i3]:
        with pytest.raises(TypeError):
            d1 - i
        with pytest.raises(TypeError):
            d1 + i


@pytest.mark.parametrize(('prop', 'expected_unit'), [
    ('nanoseconds', 'ns'),
    ('microseconds', 'us'),
    ('milliseconds', 'ms'),
    ('seconds', 's'),
    ('minutes', 'm'),
    ('hours', 'h'),
    ('days', 'D'),
    ('weeks', 'W')
])
def test_interval_properties(prop, expected_unit):
    i = api.interval(seconds=3600)
    assert getattr(i, prop).type().unit == expected_unit


@pytest.mark.parametrize('interval', [
    api.interval(years=1),
    api.interval(quarters=4),
    api.interval(months=12)
])
@pytest.mark.parametrize(('prop', 'expected_unit'), [
    ('months', 'M'),
    ('quarters', 'Q'),
    ('years', 'Y')
])
def test_interval_monthly_properties(interval, prop, expected_unit):
    assert getattr(interval, prop).type().unit == expected_unit


@pytest.mark.parametrize(('interval', 'prop'), [
    (api.interval(hours=48), 'months'),
    (api.interval(years=2), 'seconds'),
    (api.interval(quarters=1), 'weeks')
])
def test_unsupported_properties(interval, prop):
    with pytest.raises(ValueError):
        getattr(interval, prop)


@pytest.mark.parametrize('column', [
    'a', 'b', 'c', 'd'  # integer columns
])
@pytest.mark.parametrize('unit', [
    'Y', 'Q', 'M', 'D', 'W',
    'h', 'm', 's', 'ms', 'us', 'ns'
])
def test_integer_to_interval(column, unit, table):
    c = table[column]
    i = c.to_interval(unit)
    assert isinstance(i, ir.IntervalColumn)
    assert i.type().value_type == c.type()
    assert i.type().unit == unit


@pytest.mark.parametrize('unit', [
    'Y', 'Q', 'M', 'D', 'W',
    'h', 'm', 's', 'ms', 'us', 'ns'
])
@pytest.mark.parametrize('operands', [
    lambda t, u: (api.interval(3, unit=u), api.interval(2, unit=u)),
    lambda t, u: (api.interval(3, unit=u), api.interval(3, unit=u)),
    lambda t, u: (t.c.to_interval(unit=u), api.interval(2, unit=u)),
    lambda t, u: (t.c.to_interval(unit=u), t.d.to_interval(unit=u))
])
@pytest.mark.parametrize('operator', [
    operator.eq,
    operator.ne,
    operator.ge,
    operator.gt,
    operator.le,
    operator.lt
], ids=lambda op: op.__name__)
def test_interval_comparisons(unit, operands, operator, table):
    a, b = operands(table, unit)
    expr = operator(a, b)

    assert isinstance(a, ir.IntervalValue)
    assert isinstance(b, ir.IntervalValue)
    assert isinstance(expr, ir.BooleanValue)


@pytest.mark.parametrize('operands', [
    lambda t: (api.date('2016-01-01'), api.date('2016-02-02')),
    lambda t: (t.j, api.date('2016-01-01')),
    lambda t: (api.date('2016-01-01'), t.j),
    lambda t: (t.j, t.i.date()),
    lambda t: (t.i.date(), t.j)
], ids=[
    'literal-literal',
    'column-literal',
    'literal-column',
    'column-casted',
    'casted-column'
])
@pytest.mark.parametrize('interval', [
    lambda t: api.interval(years=4),
    lambda t: api.interval(quarters=4),
    lambda t: api.interval(months=3),
    lambda t: api.interval(weeks=2),
    lambda t: api.interval(days=1),
    lambda t: t.c.to_interval(unit='Y'),
    lambda t: t.c.to_interval(unit='M'),
    lambda t: t.c.to_interval(unit='W'),
    lambda t: t.c.to_interval(unit='D'),
], ids=[
    'years',
    'quarters',
    'months',
    'weeks',
    'days',
    'to-years',
    'to-months',
    'to-weeks',
    'to-days'
])
@pytest.mark.parametrize('arithmetic', [
    lambda a, i: a - i,
    lambda a, i: a + i,
    lambda a, i: i + a
], ids=[
    'subtract',
    'radd',
    'add'
])
@pytest.mark.parametrize('operator', [
    operator.eq,
    operator.ne,
    operator.ge,
    operator.gt,
    operator.le,
    operator.lt
], ids=lambda op: op.__name__)
def test_complex_date_comparisons(operands, interval, arithmetic, operator,
                                  table):
    (a, b), i = operands(table), interval(table)

    a_ = arithmetic(a, i)
    expr = operator(a_, b)

    assert isinstance(a, ir.DateValue)
    assert isinstance(b, ir.DateValue)
    assert isinstance(a_, ir.DateValue)
    assert isinstance(i, ir.IntervalValue)
    assert isinstance(expr, ir.BooleanValue)


def test_interval_column_name(table):
    c = table.i
    expr = (c - c).name('foo')
    assert expr._name == 'foo'


@pytest.mark.parametrize('operand', [
    lambda t: api.timestamp(datetime.datetime.now()),
    lambda t: t.i
], ids=[
    'column',
    'literal'
])
@pytest.mark.parametrize('unit', [
    'Y', 'y', 'year', 'YEAR', 'YYYY', 'SYYYY', 'YYY', 'YY',
    'Q', 'q', 'quarter', 'QUARTER',
    'M', 'month', 'MONTH',
    'w', 'W', 'week', 'WEEK',
    'd', 'J', 'day', 'DAY',
    'h', 'H', 'HH24', 'hour', 'HOUR',
    'm', 'MI', 'minute', 'MINUTE',
    's', 'second', 'SECOND',
    'ms', 'millisecond', 'MILLISECOND',
    'us', 'microsecond', 'MICROSECOND',
    'ns', 'nanosecond', 'NANOSECOND'
])
def test_timestamp_truncate(table, operand, unit):
    expr = operand(table).truncate(unit)
    assert isinstance(expr, ir.TimestampValue)
    assert isinstance(expr.op(), ops.TimestampTruncate)


@pytest.mark.parametrize('operand', [
    lambda t: api.date('2018-01-01'),
    lambda t: t.j,
])
@pytest.mark.parametrize('unit', [
    'Y', 'Q', 'M', 'D', 'W',
])
def test_date_truncate(table, operand, unit):
    expr = operand(table).truncate(unit)
    assert isinstance(expr, ir.DateValue)
    assert isinstance(expr.op(), ops.DateTruncate)


@pytest.mark.parametrize('operand', [
    lambda t: api.time('18:00'),
    lambda t: t.k
])
@pytest.mark.parametrize('unit', [
    'h', 'm', 's', 'ms', 'us', 'ns'
])
def test_time_truncate(table, operand, unit):
    expr = operand(table).truncate(unit)
    assert isinstance(expr, ir.TimeValue)
    assert isinstance(expr.op(), ops.TimeTruncate)



import pytest
import ibis

from pandas.util import testing as tm
pa = pytest.importorskip('pyarrow')
import pyarrow.parquet as pq  # noqa: E402

from ibis.file.parquet import ParquetClient, ParquetTable  # noqa: E402
from ibis.file.client import FileDatabase, execute  # noqa: E402


@pytest.fixture
def transformed(parquet):

    closes = parquet.pq.close
    opens = parquet.pq.open

    t = opens.inner_join(closes, ['time', 'ticker'])
    t = t[opens, closes.close]
    t = t.mutate(avg=(t.open + t.close) / 2)
    t = t[['time', 'ticker', 'avg']]
    return t


def test_creation(parquet):
    # we have existing files in our dir

    d = parquet.client.root
    assert len(list(d.iterdir())) == 1

    pqd = d / 'pq'
    assert len(list(pqd.iterdir())) == 2

    assert len(pq.read_table(str(pqd / 'open.parquet'))) == 50
    assert len(pq.read_table(str(pqd / 'close.parquet'))) == 50


def test_client(tmpdir, data):

    # construct with a path to a file
    d = tmpdir / 'pq'
    d.mkdir()

    for k, v in data.items():
        f = d / "{}.parquet".format(k)
        table = pa.Table.from_pandas(v)
        pq.write_table(table, str(f))

    c = ParquetClient(tmpdir)
    assert c.list_databases() == ['pq']
    assert c.database().pq.list_tables() == ['close', 'open']


def test_navigation(parquet):

    # directory navigation
    assert isinstance(parquet, FileDatabase)
    result = dir(parquet)
    assert result == ['pq']

    d = parquet.pq
    assert isinstance(d, FileDatabase)
    result = dir(d)
    assert result == ['close', 'open']

    result = d.list_tables()
    assert result == ['close', 'open']

    opens = d.open
    assert isinstance(opens.op(), ParquetTable)

    closes = d.close
    assert isinstance(closes.op(), ParquetTable)


def test_read(parquet, data):

    closes = parquet.pq.close
    assert str(closes) is not None

    result = closes.execute()
    expected = data['close']
    tm.assert_frame_equal(result, expected)

    result = execute(closes)
    tm.assert_frame_equal(result, expected)


def test_write(transformed, tmpdir):
    t = transformed
    expected = execute(t)

    tpath = tmpdir / 'new_dir'
    tpath.mkdir()
    path = tpath / 'foo.parquet'

    assert not path.exists()
    t = transformed[['time', 'ticker', 'avg']]
    c = ibis.parquet.connect(tpath)
    c.insert('foo.parquet', t)
    execute(t)
    assert path.exists()

    # readback
    c = ParquetClient(str(tpath)).database()
    result = c.list_databases()
    assert result == []

    result = c.foo.execute()
    tm.assert_frame_equal(result, expected)
    path = tpath / 'foo.parquet'
    assert path.exists()



import enum
import pytest

import six

import ibis
from ibis.common import IbisTypeError
import ibis.expr.operations as ops
import ibis.expr.types as ir
import ibis.expr.datatypes as dt
from ibis.expr import rules


class MyExpr(ir.Expr):
    pass


def test_enum_validator():
    class Foo(enum.Enum):
        a = 1
        b = 2

    class Bar(enum.Enum):
        a = 1
        b = 2

    class MyOp(ops.Node):

        input_type = [rules.enum(Foo, name='value')]

        def __init__(self, value):
            super(MyOp, self).__init__([value])

        def output_type(self):
            return MyExpr

    assert MyOp(2) is not None
    assert MyOp(Foo.b) is not None

    with pytest.raises(IbisTypeError):
        MyOp(3)

    with pytest.raises(IbisTypeError):
        MyOp(Bar.a)

    op = MyOp(Foo.a)
    assert op._validate_args(op.args) == [Foo.a]

    op = MyOp(2)
    assert op._validate_args(op.args) == [Foo.b]


def test_duplicate_enum():
    enum = pytest.importorskip('enum')

    class Dup(enum.Enum):
        a = 1
        b = 1
        c = 2

    class MyOp(ops.Node):

        input_type = [rules.enum(Dup, name='value')]

        def __init__(self, value):
            super(MyOp, self).__init__([value])

        def output_type(self):
            return MyExpr

    with pytest.raises(IbisTypeError):
        MyOp(1)

    assert MyOp(2) is not None


@pytest.mark.parametrize(
    ['options', 'expected_case'],
    [
        (['FOO', 'BAR', 'BAZ'], str.upper),
        (['Foo', 'Bar', 'Baz'], str.upper),  # default is upper
        (['foo', 'bar', 'BAZ'], str.lower),  # majority wins
        (['foo', 'bar', 'Baz'], str.lower),
        (['FOO', 'BAR', 'bAz'], str.upper),
        (['FOO', 'BAR', 'baz'], str.upper),
    ]
)
@pytest.mark.parametrize(
    'option',
    ['foo', 'Foo', 'fOo', 'FOo', 'foO', 'FoO', 'fOO', 'FOO',
     'bar', 'Bar', 'bAr', 'BAr', 'baR', 'BaR', 'bAR', 'BAR',
     'baz', 'Baz', 'bAz', 'BAz', 'baZ', 'BaZ', 'baZ', 'BAZ'],
)
def test_string_options_case_insensitive(options, expected_case, option):
    class MyOp(ops.Node):

        input_type = [
            rules.string_options(options, case_sensitive=False, name='value')
        ]

        def __init__(self, value):
            super(MyOp, self).__init__([value])

        def output_type(self):
            return MyExpr

    op = MyOp(option)
    assert op._validate_args(op.args) == [expected_case(option)]


def test_argument_docstring():
    doc = 'A wonderful integer'

    class MyExpr(ir.Expr):
        pass

    class MyOp(ops.ValueOp):

        input_type = [rules.integer(name='foo', doc=doc)]

        def output_type(self):
            return MyExpr

    op = MyOp(1)
    assert type(op).foo.__doc__ == doc


def test_scalar_value_type():

    class MyOp(ops.ValueOp):

        input_type = [rules.scalar(value_type=rules.number)]
        output_type = rules.double

    with pytest.raises(IbisTypeError):
        MyOp('a')

    assert MyOp(1).args[0].equals(ibis.literal(1))
    assert MyOp(1.42).args[0].equals(ibis.literal(1.42))


def test_array_rule():

    class MyOp(ops.ValueOp):

        input_type = [rules.array(dt.double, name='value')]
        output_type = rules.type_of_arg(0)

    raw_value = [1.0, 2.0, 3.0]
    op = MyOp(raw_value)
    result = op.value
    expected = ibis.literal(raw_value)
    assert result.equals(expected)


def test_scalar_default_arg():
    class MyOp(ops.ValueOp):

        input_type = [
            rules.scalar(
                value_type=dt.boolean,
                optional=True,
                default=False,
                name='value'
            )
        ]
        output_type = rules.type_of_arg(0)

    op = MyOp()
    assert op.value.equals(ibis.literal(False))

    op = MyOp(True)
    assert op.value.equals(ibis.literal(True))


def test_custom_as_value_expr():

    def custom_as_value_expr(o):
        if isinstance(o, int):
            return ibis.literal(o + 2)
        return ibis.literal(o)

    class MyOp(ops.ValueOp):
        input_type = [
            rules.list_of(
                rules.integer(as_value_expr=custom_as_value_expr),
                name='bar',
            )
        ]

    result = MyOp([2])
    bar = result.bar
    assert len(bar) == 1

    bar_value, = bar
    assert bar_value.equals(ibis.literal(4))


def test_custom_list_of_as_value_expr():

    class MyList(list):
        pass

    class MyEnum(enum.Enum):
        A = 1
        B = 2

    class MyEnum2(enum.Enum):
        A = 1
        B = '2'

    def custom_as_value_expr(o):
        if o and all(isinstance(el.value, six.integer_types) for el in o):
            return MyList(o)
        return o

    class MyOp(ops.ValueOp):

        input_type = [
            rules.list_of(
                rules.enum(MyEnum),
                name='one',
                as_value_expr=custom_as_value_expr
            ),
            rules.list_of(
                rules.enum(MyEnum2),
                name='two',
                as_value_expr=custom_as_value_expr
            ),
        ]

    result = MyOp([MyEnum.A, MyEnum.B], [])
    assert isinstance(result.one, MyList)
    assert result.one == [MyEnum.A, MyEnum.B]
    assert result.two == []

    result = MyOp([MyEnum.A, MyEnum.B], [MyEnum2.B])
    assert isinstance(result.one, MyList)
    assert not isinstance(result.two, MyList)
    assert result.one == [MyEnum.A, MyEnum.B]
    assert result.two == [MyEnum2.B]


def check_op_input(Op, schema, raises):
    schema = ibis.Schema.from_tuples(schema)
    table = ibis.table(schema)
    if not raises:
        assert Op(table).table.equals(table)
    else:
        with pytest.raises(IbisTypeError):
            Op(table)


@pytest.mark.parametrize(
    'schema',
    [[('group', dt.int64),
      ('value', dt.double)],

     [('group', dt.int64),
      ('value', dt.double),
      ('value2', dt.double)]])
def test_table_with_schema(schema):
    class MyOp(ops.ValueOp):
        input_type = [
            rules.table(
                name='table',
                schema=rules.table.with_column_subset(
                    rules.column(name='group', value_type=rules.number),
                    rules.column(name='value', value_type=rules.number)
                ))]
        output_type = rules.type_of_arg(0)

    schema = ibis.Schema.from_tuples(schema)
    table = ibis.table(schema)
    MyOp(table)


@pytest.mark.parametrize(
    'schema',
    [[('group', dt.int64),
      ('value', dt.timestamp)]])
def test_table_with_schema_invalid(schema):
    class MyOp(ops.ValueOp):
        input_type = [
            rules.table(
                name='table',
                schema=rules.table.with_column_subset(
                    rules.column(name='group', value_type=rules.number),
                    rules.column(name='value', value_type=rules.number)
                ))]
        output_type = rules.type_of_arg(0)

    schema = ibis.Schema.from_tuples(schema)
    table = ibis.table(schema)
    with pytest.raises(IbisTypeError):
        MyOp(table)


@pytest.mark.parametrize(
    'schema',
    [[('group', dt.int64),
      ('value', dt.double)],

     [('group', dt.int64),
      ('value', dt.double),
      ('value2', dt.double)],

     [('group', dt.int64),
      ('value', dt.double),
      ('value3', dt.timestamp)],

     [('group', dt.int64),
      ('value', dt.double),
      ('value2', dt.double),
      ('value3', dt.double)],

     [('group', dt.int64),
      ('value', dt.double),
      ('value2', dt.double),
      ('value3', dt.timestamp)]])
def test_table_with_schema_optional(schema):
    class MyOp(ops.ValueOp):
        input_type = [
            rules.table(
                name='table',
                schema=rules.table.with_column_subset(
                    rules.column(name='group', value_type=rules.number),
                    rules.column(name='value', value_type=rules.number),
                    rules.column(name='value2', value_type=rules.number,
                                 optional=True)))]
        output_type = rules.type_of_arg(0)

    schema = ibis.Schema.from_tuples(schema)
    table = ibis.table(schema)
    MyOp(table)


@pytest.mark.parametrize(
    'schema',
    [[('group', dt.int64),
      ('value', dt.double),
      ('value2', dt.timestamp)]])
def test_table_with_schema_optional_invalid(schema):
    class MyOp(ops.ValueOp):
        input_type = [
            rules.table(
                name='table',
                schema=rules.table.with_column_subset(
                    rules.column(name='group', value_type=rules.number),
                    rules.column(name='value', value_type=rules.number),
                    rules.column(name='value2', value_type=rules.number,
                                 optional=True)))]
        output_type = rules.type_of_arg(0)

    schema = ibis.Schema.from_tuples(schema)
    table = ibis.table(schema)
    with pytest.raises(IbisTypeError):
        MyOp(table)


def test_table_not_a_table():
    class MyOp(ops.ValueOp):
        input_type = [rules.table(name='table')]
        output_type = rules.type_of_arg(0)

    with pytest.raises(IbisTypeError):
        MyOp(123)


def test_table_invalid_schema_no_name():
    with pytest.raises(ValueError):
        class MyOp(ops.ValueOp):
            input_type = [rules.table(
                name='table',
                schema=rules.table.with_column_subset(
                    rules.column(value_type=rules.number)))]
            output_type = rules.type_of_arg(0)


def test_table_invalid_schema_wrong_class():
    with pytest.raises(ValueError):
        class MyOp(ops.ValueOp):
            input_type = [rules.table(
                name='table',
                schema='wrong class')]
            output_type = rules.type_of_arg(0)


def test_table_invalid_column_subset():
    with pytest.raises(ValueError):
        class MyOp(ops.ValueOp):
            input_type = [rules.table(
                name='table',
                schema=rules.table.with_column_subset('not a rule'))]
            output_type = rules.type_of_arg(0)


def test_table_custom_validator():
    class MyOp(ops.ValueOp):
        input_type = [rules.table(
            name='table',
            validator=lambda x: 1 / 0)]
        output_type = rules.type_of_arg(0)

    schema = ibis.Schema.from_tuples([('group', dt.int64)])
    table = ibis.table(schema)

    with pytest.raises(ZeroDivisionError):
        MyOp(table)



import pytest

import ibis
import ibis.tests.util as tu
from ibis.compat import parse_version


@tu.skipif_unsupported
def test_version(backend, con):
    expected_type = (type(parse_version('1.0')),
                     type(parse_version('1.0-legacy')))
    assert isinstance(con.version, expected_type)


@pytest.mark.parametrize(('expr_fn', 'expected'), [
    (lambda t: t.string_col, [('string_col', 'string')]),
    (lambda t: t[t.string_col, t.bigint_col],
     [('string_col', 'string'), ('bigint_col', 'int64')])
])
def test_query_schema(backend, con, alltypes, expr_fn, expected):
    if not hasattr(con, '_build_ast'):
        pytest.skip(
            '{} backend has no _build_ast method'.format(
                type(backend).__name__
            )
        )

    expr = expr_fn(alltypes)

    # we might need a public API for it
    ast = con._build_ast(expr, backend.make_context())
    query = con.sync_query(con, ast.queries[0])

    # test single columns
    expected = ibis.schema(expected)
    assert query.schema().equals(expected)
